#include <array>
#include <vector>

#include <dune/common/reservedvector.hh>
#include <dune/functions/functionspacebases/flatmultiindex.hh>

#include <amdis/common/FieldMatVec.hpp>
#include <amdis/common/Concepts.hpp>

using namespace AMDiS;

int main()
{
  using V = Dune::FieldVector<double,2>;
  using M = Dune::FieldMatrix<double,2,2>;

  // arithmetic concepts
  static_assert( Concepts::Addable<double,double>, "" );
  static_assert( Concepts::Addable<V,V>, "" );
  static_assert( Concepts::Addable<M,M>, "" );

  static_assert( Concepts::Subtractable<double,double>, "" );
  static_assert( Concepts::Subtractable<V,V>, "" );
  static_assert( Concepts::Subtractable<M,M>, "" );

  static_assert( Concepts::Multiplicable<double,double>, "" );
  static_assert( Concepts::Multiplicable<V,double>, "" );
  static_assert( Concepts::Multiplicable<V,V>, "" );
  static_assert( Concepts::Multiplicable<M,double>, "" );
//  static_assert( Concepts::Multiplicable<M,V>, "" ); // A*x defined but x*A is not

  static_assert( Concepts::Divisible<double,double>, "" );
  static_assert( Concepts::Divisible<V,double>, "" );
  static_assert( Concepts::Divisible<M,double>, "" );

  static_assert( Concepts::LessThanComparable<double,double>, "" );
  static_assert( Concepts::LessThanComparable<char,char>, "" );

  // Concepts::Callable
  auto f = [](V const&) { return 0.0; };
  auto b = [](V const&) { return true; };

  static_assert( Concepts::Callable<decltype(f),V>, "" );
  static_assert( Concepts::Evaluable<decltype(f),V>, "" );
  static_assert( Concepts::Functor<decltype(f),double(V)>, "" );
  static_assert( Concepts::Functor<decltype(b),bool(V)>, "" );
  static_assert( Concepts::Predicate<decltype(b),V>, "" );

  // Concepts::MultiIndex
  static_assert( Concepts::MultiIndex<Dune::ReservedVector<std::size_t,2>>, "" );
  static_assert( Concepts::MultiIndex<Dune::Functions::FlatMultiIndex<std::size_t>>, "" );
  static_assert( Concepts::MultiIndex<std::array<std::size_t,2>>, "" );
  static_assert( Concepts::MultiIndex<std::vector<std::size_t>>, "" );
}
